{ =========================================================================== }
{ Window34.inc - Multi-level windowing procedures           ver 3.4, 08-31-87 }
{                                                                             }
{ This file allows you to produce quick multi-level windows for any IBM or    }
{ compatibles in any column mode (40/80/etc.).  You should get a copy of      }
{ QWIK30.ARC or a later version to make full use of quick screen writing      }
{ procedures.  This file has been released as Shareware.                      }
{ =========================================================================== }

type
  Borders = (NoBrdr, BlankBrdr, SingleBrdr, DoubleBrdr, MixedBrdr, SolidBrdr,
             EvenSolidBrdr, ThinSolidBrdr, LhatchBrdr, MhatchBrdr,
             HhatchBrdr, UserBrdr);
  BrdrRec = record
              TL,TH,TR,LV,RV,BL,BH,BR: string[1];
            end;
  DirType = (NoDir,Up,Down,VeryTop,Top,Bottom,VeryBottom,FarLeft,Left,Right,
             FarRight,Center);
  WndwStatType = record
              WSrow,WScol,WSrows,WScols,WSWattr,WSBattr: byte;
              WSbrdr:          Borders;
              WSshadow:        DirType;
              WSlastx,WSlasty: byte;
              WSfixed:         boolean;
            end;
  BytePtr = ^byte;
  Str160  = String[160];

{ The following constants are typed so there's no need to change this file.  }
{ Just assign them new values in your main program like any other variable.  }
{ UserBrdr is also one you can use for scratch while keeping the others.     }
const
  Brdr: array [BlankBrdr..UserBrdr] of BrdrRec =
    ((TL:' ';TH:' ';TR:' ';LV:' ';RV:' ';BL:' ';BH:' ';BR:' '),  { Blank     }
     (TL:'';TH:'';TR:'';LV:'';RV:'';BL:'';BH:'';BR:''),  { Single    }
     (TL:'';TH:'';TR:'';LV:'';RV:'';BL:'';BH:'';BR:''),  { Double    }
     (TL:'';TH:'';TR:'';LV:'';RV:'';BL:'';BH:'';BR:''),  { Mixed     }
     (TL:'';TH:'';TR:'';LV:'';RV:'';BL:'';BH:'';BR:''),  { Solid     }
     (TL:'';TH:'';TR:'';LV:'';RV:'';BL:'';BH:'';BR:''),  { EvenSolid }
     (TL:'';TH:'';TR:'';LV:'';RV:'';BL:'';BH:'';BR:''),  { ThinSolid }
     (TL:'';TH:'';TR:'';LV:'';RV:'';BL:'';BH:'';BR:''),  { Lhatch    }
     (TL:'';TH:'';TR:'';LV:'';RV:'';BL:'';BH:'';BR:''),  { Mhatch    }
     (TL:'';TH:'';TR:'';LV:'';RV:'';BL:'';BH:'';BR:''),  { Hhatch    }
     (TL:' ';TH:' ';TR:' ';LV:' ';RV:' ';BL:' ';BH:' ';BR:' ')); { User      }
  ShadowEffect: DirType = NoDir;
  ZoomEffect:   boolean = false;
  ZoomDelay:    byte = 11;
  FixedWindow:  boolean = false;

var
     WndwStat: Array [0..MaxWndw] of WndwStatType;         { window stats }
     WndwPtr:  Array [1..MaxWndw] of BytePtr; { pointer to window on heap }
     LI,LIcurrent: byte;  { level index, level index for swapping windows }
     Tattr:        byte absolute Dseg:$0008;    { Turbo's attribute value }

{ =========================================================================== }
{ NAME: Attr                                               ver 3.1,  02-11-87 }
{ DESCRIPTION: Converts Turbo color constants into an attribute               }
{              and masks any accidental blink bit.                            }
{ PARAMETERS:  ForeGround - Color of text foreground                          }
{              BackGround - Color of text background                          }
{ =========================================================================== }
function Attr (Foreground,Background: byte): byte;
begin
  Attr := ((BackGround shl 4) + ForeGround) and 127;
end;

{ =========================================================================== }
{ NAME: Qbox                                               ver 3.4,  08-31-87 }
{ DESCRIPTION: Writes a window with optional border.  Since attribute         }
{              is byte, the colors should always be specified.                }
{ PARAMETERS:  See QWIK30.DOC                                                 }
{ =========================================================================== }
procedure Qbox (Row,Col,Rows,Cols,WndwAttr,BrdrAttr: byte; BrdrSel: Borders);
begin
  if (Rows>=2) and (Cols>=2) then
  begin
    if BrdrSel<>NoBrdr then
      with Brdr[BrdrSel] do
      begin
        QwriteV (     Row      ,     Col                    ,BrdrAttr,TL);
        Qfill   (     Row      ,succ(Col)     ,1     ,Cols-2,BrdrAttr,TH);
        QwriteV (     Row      ,pred(Col)+Cols              ,BrdrAttr,TR);
        Qfill   (succ(Row)     ,     Col      ,Rows-2,1     ,BrdrAttr,LV);
        Qfill   (succ(Row)     ,pred(Col)+Cols,Rows-2,1     ,BrdrAttr,RV);
        QwriteV (pred(Row)+Rows,     Col                    ,BrdrAttr,BL);
        Qfill   (pred(Row)+Rows,succ(Col)     ,1     ,Cols-2,BrdrAttr,BH);
        QwriteV (pred(Row)+Rows,pred(Col)+Cols              ,BrdrAttr,BR);
        Qfill   (succ(Row)     ,succ(Col)     ,Rows-2,Cols-2,WndwAttr,' ')
      end
    else Qfill  (Row,Col,Rows,Cols,WndwAttr,' ');
  end
end;

{ =========================================================================== }
{ NAME: InitWindow                                          ver 3.4, 08-31-87 }
{ DESCRIPTION:  Initializes the window variables.  Use this routine before    }
{               using MakeWindow, RemoveWindow, TitleWindow or ScrollWindow   }
{ PARAMETERS:                                                                 }
{       Wattr - Starting window attribute                                     }
{ =========================================================================== }
procedure InitWindow (Wattr: byte);
begin
   Qinit;                  { QWIK30.INC initialization !!!! }
   Tattr := Wattr;
   LI := 0;
   LIcurrent := 0;
   with WndwStat[LI] do
     begin
       WSrow   := 1;       { Initialize non-window zero }
       WScol   := 1;
       WSrows  := 25;
       WScols  := 80;
       WSWattr := Wattr;
       WSBattr := Wattr;
       WSbrdr  := NoBrdr;
       WSlastx := WhereX;
       WSlasty := WhereY;
       WSfixed := true;
     end;
   Qfill ( 1, 1,25,80,Wattr,' ')
end;

{ =========================================================================== }
{ NAME: HeapOK                                              ver 3.3, 04-08-87 }
{ DESCRIPTION:  Checks for enough memory on the heap used by MakeWindow.      }
{ PARAMETERS:   Bytes - number of bytes needed on the heap                    }
{ =========================================================================== }
function HeapOK (Bytes: integer): boolean;
begin
   { if memavail<0 then there's plenty of room (>512kb) }
  if (0<memavail) and (memavail<=(Bytes shr 4)) then
    begin
      writeln('Not enough Heap space!',^G^G);
      HeapOK := false
    end
  else HeapOK := true
end;

{ =========================================================================== }
{ NAME: MakeWindow                                          ver 3.4, 08-31-87 }
{ DESCRIPTION:  Creates a window on your screen.                              }
{ PARAMETERS:                                                                 }
{       Row    - First row        (1 - Screen limit)                          }
{       Col    - First column     (1 - Screen limit)                          }
{       Rows   - # of rows        (1 - Screen limit)                          }
{       Cols   - # of columns     (1 - Screen limit)                          }
{       Wattr  - Window attribute (0 - 255)                                   }
{       Battr  - Border attribute (0 - 255)                                   }
{       BrdSel - Border selection (NoBrdr - UserBrdr)                         }
{ =========================================================================== }
procedure MakeWindow (Row,Col,Rows,Cols,Wattr,Battr: byte; BrdrSel: Borders);
var Wsize,r1,r2,c1,c2,ColRatio: integer;
    OK:                         boolean;
{}procedure TurboWindow (Row,Col,Rows,Cols: byte);
  begin
    window (Col,Row,pred(Col)+Cols,pred(Row)+Rows);
    GotoRC (Row,Col)
{}end;
{}procedure ShadowFill (VertCol,HorizCol: byte);
  begin
    Qfill (succ(Row),VertCol ,pred(Rows),2   ,0,' ');
    Qfill (Row+Rows ,HorizCol,1         ,Cols,0,' ')
{}end;
begin
  if LI>=MaxWndw then WriteLn('Too many Windows!',^G^G)
  else
  begin
    case ShadowEffect of
      Left:  begin c1:=Col-2; c2:=Cols+2; r2:=succ(Rows) end;
      Right: begin c1:=Col;   c2:=Cols+2; r2:=succ(Rows) end;
      else   begin c1:=Col;   c2:=Cols;   r2:=     Rows; end;
    end;
    Wsize := r2*c2 shl 1;           { Memory size needed to store display }
    if FixedWindow then
          OK:=true
    else OK:=HeapOK(Wsize);
    if OK then
      begin
        WndwStat[LI].WSlastx := wherex;   { Store old cursor coordinates }
        WndwStat[LI].WSlasty := wherey;
        LI := succ(LI);        { Go to next window level }
        LIcurrent := LI;       { Current level = Top level }
        Tattr := Wattr;
        with WndwStat[LI] do
          begin
            WSrow   := Row;      { Store all variables for this window }
            WScol   := Col;
            WSrows  := Rows;
            WScols  := Cols;
            WSWattr := Wattr;
            WSBattr := Battr;
            WSbrdr  := BrdrSel;
            WSshadow:= ShadowEffect;
            WSfixed := FixedWindow;
          end;
        if not FixedWindow then
          begin
            GetMem (WndwPtr[LI],Wsize);   { Get heap to store old display }
            QstoreToMem (Row,c1,r2,c2,WndwPtr[LI]^);
          end;
        if ZoomEffect then
          begin
            r1 := row+     (rows shr 1);
            r2 := row+rows-(rows shr 1);
            c1 := col+     (cols shr 1);
            c2 := col+cols-(cols shr 1);
            ColRatio := succ(cols div rows);
            if ColRatio>4 then ColRatio:=4;
            repeat
              if r1>row        then r1:=pred(r1);
              if r2<(row+rows) then r2:=succ(r2);
              if c1>col        then c1:=c1-ColRatio;
              if c1<col        then c1:=col;
              if c2<(col+cols) then c2:=c2+ColRatio;
              if c2>(col+cols) then c2:=col+cols;
              Qbox (r1,c1,r2-r1,c2-c1,Tattr,Battr,BrdrSel);
              if Qwait=false then delay (ZoomDelay);
            until  (c1=col) and (c2=col+cols) and (r1=row) and (r2=row+rows)
          end
        else Qbox (Row,Col,Rows,Cols,Wattr,Battr,BrdrSel);
        case ShadowEffect of
          Left:  ShadowFill (Col-2   ,Col-2);
          Right: ShadowFill (Col+Cols,Col+2);
        end;
        if BrdrSel=NoBrdr then
             TurboWindow (     Row ,     Col ,Rows  ,Cols  )
        else TurboWindow (succ(Row),succ(Col),Rows-2,Cols-2);
        FixedWindow:=false;
      end; { Continue }
  end      { if LI }
end;

{ =========================================================================== }
{ NAME: RemoveWindow                                        ver 3.4, 08-31-87 }
{ DESCRIPTION: Remove the last window created from the screen.  To            }
{              get back to the original screen, there must be as many         }
{              RemoveWindow(s) as there are MakeWindow(s).                    }
{ =========================================================================== }
procedure RemoveWindow;
{}procedure TurboWindow (Row,Col,Rows,Cols: byte);
  begin
    window (Col,Row,pred(Col)+Cols,pred(Row)+Rows);
    with WndwStat[LI] do
      GotoRC (pred(Row)+WSlasty,pred(Col)+WSlastx)
{}end;
{}procedure RestoreScr (Row,Col,Rows,Cols: byte);
  begin
    QstoreToScr (Row,Col,Rows,Cols,WndwPtr[LI]^);
    FreeMem (WndwPtr[LI],Rows*Cols shl 1)
{}end;
begin
  if LI=0 then WriteLn ('No Window To Remove!',^G^G)
  else
  begin
    with WndwStat[LI] do
      if not WSfixed then
        case WSshadow of
          Left:  RestoreScr (WSrow,WScol-2,succ(WSrows),WScols+2);
          Right: RestoreScr (WSrow,WScol  ,succ(WSrows),WScols+2);
          else   RestoreScr (WSrow,WScol  ,     WSrows ,WScols  );
        end;
    LI := pred(LI);        { Go to next lower level }
    LIcurrent := LI;       { Current level = Top level }
    with WndwStat[LI] do
      begin
        Tattr:= WSWattr;
        if WSbrdr=NoBrdr then
             TurboWindow (WSrow      ,WScol      ,WSrows  ,WScols  )
        else TurboWindow (succ(WSrow),succ(WScol),WSrows-2,WScols-2);
      end
  end     { if LI }
end;

{ =========================================================================== }
{ NAME: TitleWindow                                         ver 3.4, 08-31-87 }
{ DESCRIPTION: Places a centered title in the top border of a window          }
{ PARAMETERS:  Justify - justification of the title                           }
{              Title - Optional title of the window                           }
{ =========================================================================== }
procedure TitleWindow (Justify: DirType; title: Str160);
begin
  with WndwStat[LI] do
    case Justify of
      Left   : QwriteV  (WSrow,WScol+2, -1,title);
      Center : QwriteCV (WSrow,WScol,pred(WScol)+WScols, -1,title);
      Right  : QwriteV  (WSrow,WScol+WScols-Length(Title)-2, -1,title);
    end;
end;

{ =========================================================================== }
{ NAME: ScrollWindow                                        ver 3.4, 08-31-87 }
{ DESCRIPTION: Scrolls a number of rows in a window.  Using a little          }
{              thought, you can see how this is better than the InsLine       }
{              and DelLine procedures.  Flicker-free and works on any page.   }
{ PARAMETERS:  RowBegin,RowEnd - Rows to be affected                          }
{              Dir             - 'Up' or 'Down'                               }
{ =========================================================================== }
procedure ScrollWindow (RowBegin,RowEnd: byte; Dir: DirType);
var BrdrWidth,R,C,Rs,Cs: byte;
{}procedure Qscroll (MemRowBegin,ScrRowBegin,FillRow: byte);
  var Temp:     BytePtr;
      TempSize: integer;
  begin
    TempSize := Rs*Cs;
    if HeapOK(TempSize) then
    begin
      GetMem      (Temp,TempSize);
      QstoreToMem (MemRowBegin,C,Rs,Cs,Temp^);
      QstoreToScr (ScrRowBegin,C,Rs,Cs,Temp^);
      Qfill       (FillRow    ,C, 1,Cs,WndwStat[LI].WSWattr,' ');
      FreeMem     (Temp,TempSize)
    end
{}end;
begin
  with WndwStat[LI] do
  begin
    if WSbrdr=NoBrdr then
         BrdrWidth:=0
    else BrdrWidth:=1;
    R  := WSrow+BrdrWidth+pred(RowBegin);
    C  := WScol+BrdrWidth;
    Rs := RowEnd-RowBegin;
    Cs := WScols-(BrdrWidth shl 1);
    case Dir of
      Up:   Qscroll (succ(R),     R ,R+Rs);
      Down: Qscroll (     R ,succ(R),R   );
    end
  end
end;
